# Copyright (c) 2005-2010 Infrae. All rights reserved.
# See also LICENSE.txt

from datetime import datetime

from AccessControl import ClassSecurityInfo
from AccessControl.Permissions import access_contents_information, view_management_screens
from App.class_init import InitializeClass
from Persistence import Persistent
import Acquisition

from Products.OAICore.helpers import keysSortedByValue
from Products.OAICore import schemaRegistry


class BaseQuery:
    """Query base class.
    """

    security = ClassSecurityInfo()

    # default values for old query version.
    _sort = None
    _sort_order = None

    def __init__(self, service=None, metadata_format=None, name=None, filters=None, setSpec=None):
        self._service_path = None
        self._metadata_format = metadata_format
        self._schema_name = name
        if filters is None:
            # Use None in default to prevent unwanted behaviors
            filters = {}
        self._filters = filters
        self._setSpec = setSpec
        self._sort = None
        self._sort_order = None
        if service is not None:
            self.setService(service)

    # ACCESSORS

    security.declareProtected(view_management_screens, 'setService')
    def setService(self, service):
        """Set the OAI service the query is to use.
        """
        self._service_path = service.getPhysicalPath()

    security.declareProtected(view_management_screens, 'getService')
    def getService(self):
        """Get the OAI service the query is to use.
        """
        assert self._service_path is not None, "Query service is not configured."
        return self.unrestrictedTraverse(self._service_path)

    security.declareProtected(view_management_screens, 'getStorage')
    def getStorage(self):
        """Get the OAI storage.
        """
        assert self._metadata_format is not None, "Query metadata prefix is not configured."
        return self.getService().getStorage(self._metadata_format)

    security.declareProtected(view_management_screens, 'getMetadataFormat')
    def getMetadataFormat(self):
        return self._metadata_format

    security.declareProtected(view_management_screens, 'getSchemaName')
    def getSchemaName(self):
        return self._schema_name

    security.declarePrivate('getCatalog')
    def getCatalog(self):
        return self.getService().getCatalog(self._metadata_format)

    security.declarePrivate('getSchema')
    def getSchema(self):
        return schemaRegistry.getSchema(
            self._metadata_format, self._schema_name)

    security.declareProtected(access_contents_information, 'getSet')
    def getSet(self):
        return self._setSpec

    security.declareProtected(access_contents_information, 'getFields')
    def getFields(self):
        return self.getSchema().getFields()

    security.declareProtected(
        access_contents_information, 'getManageableFields')
    def getManageableFields(self):
        return self.getSchema().getManageableFields()

    security.declareProtected(
        access_contents_information, 'getInitialSortOn')
    def getInitialSortOn(self):
        return self._sort or self.getSchema().getInitialSortOn()

    security.declareProtected(
        access_contents_information, 'getInitialSortOrder')
    def getInitialSortOrder(self):
        return self._sort_order or self.getSchema().getInitialSortOrder()


    security.declareProtected(access_contents_information, 'getColumns')
    def getColumns(self):
        return self.getSchema().getResultColumns()

    security.declarePrivate('getFilters')
    def getFilters(self):
        return self._filters

    security.declareProtected(access_contents_information, 'queryResults')
    def queryResults(self, query):
        catalog = self.getCatalog()
        query.setdefault('sort_order', self.getInitialSortOrder())
        query.setdefault('sort_on', self.getInitialSortOn())
        if self._setSpec is not None:
            query['setSpec'] = self._setSpec
        filters = self.getFilters()
        for field in self.getManageableFields():
            field.updateQuery(query, filters)
        self.restrictQuery(query)
        return catalog.searchResults(**query)

    security.declarePrivate('restrictQuery')
    def restrictQuery(self, query):
        """ Just before the catalog is queried, this hook is called.

        Override it in your sub class to enforce extra restrictions on the
        query.
        """
        pass

InitializeClass(BaseQuery)


class ManageableQuery(BaseQuery):

    security = ClassSecurityInfo()

    # MANIPULATORS

    security.declareProtected(view_management_screens, 'setFilters')
    def setFilters(self, filters):
        """Set the filters dictionary (Filter results to be returned on
        values of fields in self._filters.)
        """
        self._filters = {}
        for key in filters.keys():
            if filters[key]:
                self._filters[key] = filters[key]

    security.declareProtected(view_management_screens, 'setSet')
    def setSet(self, setSpec):
        """Only return items in particular set.
        """
        # XXX I think the 'and setSpec != ['none']' can be phased out,
        # at least SilvaOAI No longer needs it
        if setSpec and setSpec != ['none']:
            self._setSpec = setSpec
        else:
            self._setSpec = None

    security.declareProtected(view_management_screens, 'setSchema')
    def setSchema(self, format, name):
        """Set the schema.
        """
        self._metadata_format = format
        self._schema_name = name
        self._sort = self.getSchema().getInitialSortOn()

    security.declareProtected(view_management_screens, 'setSort')
    def setSort(self, sort_on):
        """Set the index to sort on."""
        self._sort = sort_on

    security.declareProtected(view_management_screens, 'setSortOrder')
    def setSortOrder(self, sort_order):
        """Set the sort order.
        """
        self._sort_order = sort_order

    security.declarePrivate('addFilterFields')
    def addFilterFields(self, form):
        catalog = self.getCatalog()
        filters = self.getFilters()
        for field in self.getManageableFields():
            field.addEditField(form, filters, catalog)

    security.declarePrivate('addSortField')
    def addSortField(self, form):
        form.manage_addField('sort_on', 'Sort On', 'ListField')
        values = form.get_field('sort_on').values
        sortables = self.getSchema().getMultiSortables()
        values['items'] = [(', '.join(
            sortable.id.split('_')[1:]), sortable.id) for sortable in sortables]
        sort_on = self.getInitialSortOn()
        if sort_on:
            values['default'] = sort_on

    security.declarePrivate('addSorterField')
    def addSortOrderField(self, form):
        form.manage_addField('sort_order',
                             'Sort Order',
                             'ListField')
        values = form.get_field('sort_order').values
        values['size'] = 1
        values['items'] = ['ascending', 'descending']
        sort_order = self.getInitialSortOrder()
        if sort_order is not None:
            values['default'] = sort_order
        else:
            values['default'] = 'ascending'

    security.declarePrivate('addSetField')
    def addSetField(self, form):
        form.manage_addField('set', 'Set', 'MultiCheckBoxField')
        values = form.get_field('set').values
        values['size'] = 5
        values['required'] = 0
        values['unicode'] = 1
        filters = self.getFilters()
        if filters.has_key('set'):
            values['default'] = self._filters['set']
        service = self.getService()
        toplevel_sets = service.getSubSets(None)
        # set up the 'Set' form field
        values['items'] = result = []
        for key in keysSortedByValue(toplevel_sets):
            result.append((toplevel_sets[key], key))
            sets = service.getSubSets(key)
            for subkey in keysSortedByValue(sets):
                result.append(('... %s' % sets[subkey], subkey))

    security.declareProtected(view_management_screens, 'updateManageForm')
    def updateManageForm(self, form):
        """
        Override this method to ad custom fields to the manage form.
        Remember to call addFilterFields (and optionally addSetField
        and/or addSortField) when you do.
        """
        self.addFilterFields(form)
        self.addSortField(form)
        try:
            func = self.addExtraManageFields
        except AttributeError:
            pass
        else:
            func(form)

InitializeClass(ManageableQuery)


class UniqueValues(Acquisition.Implicit, Persistent):

    """Cache for the unique values for catalog 'metadata' columns.
    """

    def __init__(self):
        self._timestamp = datetime(1970, 1, 1) # a long time ago, to
                                               # start with an invalid
                                               # cache
        self._unique_values_for_column = {}

    def _uniqueValuesForColumn(self, column, query):
        # We cannot use the uniqueValuesFor method on the catalog, because
        # we need to apply the filters.
        unique = {}
        for brain in self.queryResults(query):
            values = brain[column]
            if values is None:
                continue
            if type([]) is not type(values):
                values = [values]
            for value in values:
                unique[value] = None
        unique_values = unique.keys()
        unique_values.sort()
        return unique_values

    def __call__(self, column, query=None):
        # First check the cache timestamp against the
        # last time the service harvested.
        service_timestamp = self.getService().getDatestamp()
        if service_timestamp > self._timestamp:
            self._unique_values_for_column = {}
            self._timestamp = datetime.utcnow()
        # See if it is cached already
        values = self._unique_values_for_column.get(column, None)
        if values is not None:
            return values
        # No, it isn't, so compute (and cache) it now
        if query is None:
            query = {}
        values = self._uniqueValuesForColumn(column, query)
        self._unique_values_for_column[column] = values
        return values


class SearchQuery(ManageableQuery):

    security = ClassSecurityInfo()

    def __init__(self, *args):
        ManageableQuery.__init__(self, *args)
        self._resetCache()

    def _resetCache(self):
        self._unique_values = UniqueValues()

    security.declareProtected(view_management_screens, 'setFilters')
    def setFilters(self, filters):
        """Set the filters dictionary (Filter results to be returned on
        values of fields in self._filters.)
        """
        self._resetCache()
        ManageableQuery.setFilters(self, filters)

    security.declareProtected(access_contents_information, 'getSearchables')
    def getSearchables(self):
        return self.getSchema().getUserSearchables()

    security.declareProtected(
        access_contents_information, 'updateSearchForm')
    def updateSearchForm(self, form):
        pass

    security.declarePrivate('addSearchFields')
    def addSearchFields(self, form):
        searchables = self.getSearchables()
        filters = self.getFilters()
        #XXX change so that the fields are looked up by id, to ensure
        #the order in user_searchables is preserved.
        for field in self.getFields():
            id = field.getId()
            # XXX Users can not search on filtered fields, which sucks:
            # let's change this, once the (z3) Catalog allows us to do
            # easy intersections.
            if id not in searchables or filters.has_key(id):
                continue
            unique_values = []
            if field.getIndexType() != 'KeywordIndex':
                continue
            for value in self.getUniqueValues(id):
                unique_values.append((value[:60], value))
            field.addPublicField(form, unique_values)

    security.declarePrivate('addFulltextSearchField')
    def addFulltextSearchField(self, form):
        form.manage_addField('fulltext', 'Fulltext Search', 'StringField')
        values = form.get_field('fulltext').values
        values['required'] = 0
        values['unicode'] = 1

    security.declareProtected(access_contents_information, 'getUniqueValues')
    def getUniqueValues(self, column):
        return self._unique_values(column)

InitializeClass(SearchQuery)
